/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#include <el/ext.h>

#include <el/comp/ext-opencl.h>

#include "miner.h"
#include "bitcoin-sha256.h"

namespace Coin {

class OpenclMiner;

class OpenclTask : public GpuTask {
	typedef GpuTask base;
public:
	OpenclMiner& m_clMiner;

	Cl::Buffer Buffer;
	Cl::Event m_ev;
	int m_buf[512];
	int m_bufTest[512];

	OpenclTask(OpenclMiner& clMiner);
	void Prepare(const BitcoinWorkData& wd, WorkerThreadBase *pwt) override;
	void Run() override;
	bool IsComplete() override;
	void Wait() override;
	int ParseResults() override;
};

class OpenclMiner : public GpuMiner {
	typedef GpuMiner base;
public:
	Cl::Device Device;
	Cl::Context Ctx;
	Cl::Program Prog;
	Cl::Kernel Kern;	
	Cl::CommandQueue Queue;

	OpenclMiner(BitcoinMiner& miner, const Cl::Device& dev, const ConstBuf& binary);

//!!!	void InitBeforeRun() override;
	ptr<GpuTask> CreateTask() override { return new OpenclTask(_self); }
	int Run(const BitcoinWorkData& wd, WorkerThreadBase *pwt) override;
private:
};

static CEvent s_ev;

void CL_CALLBACK EventNotify(cl_event ev, cl_int status, void *userData) {
	s_ev.Set();
}

OpenclTask::OpenclTask(OpenclMiner& clMiner)
	:	base(clMiner)
	,	m_clMiner(clMiner)
{
	ZeroStruct(m_buf);
	Buffer = m_clMiner.Ctx.CreateBuffer(m_buf, sizeof m_buf, CL_MEM_WRITE_ONLY|CL_MEM_USE_HOST_PTR);
}

void OpenclTask::Prepare(const BitcoinWorkData& wd, WorkerThreadBase *pwt) {
	m_pwt = pwt;
	WorkData = wd;
	m_nparOrig = WorkData.LastNonce-WorkData.FirstNonce+1,
	m_npar = m_nparOrig; //!!!?

	PrepareData(WorkData.Midstate.constData(), WorkData.Data.constData()+64, WorkData.Hash1.constData());
	m_w[3] = wd.FirstNonce;

}

void OpenclTask::Run() {
	if (!m_clMiner.Queue.Valid()) {
		m_clMiner.Queue = m_clMiner.Ctx.CreateCommandQueue(m_clMiner.Device);
	}

	for (int i=0; i<8; ++i)
		m_clMiner.Kern.SetArg(i, m_midstate[i]);

	m_clMiner.Kern.SetArg(8, m_midstate_after_3[4]);	// B1
	m_clMiner.Kern.SetArg(9, m_midstate_after_3[5]);	// C1
	m_clMiner.Kern.SetArg(10, m_midstate_after_3[6]);	// D1
	m_clMiner.Kern.SetArg(11, m_midstate_after_3[0]);	// F1
	m_clMiner.Kern.SetArg(12, m_midstate_after_3[1]);	// G1
	m_clMiner.Kern.SetArg(13, m_midstate_after_3[2]);	// H1
	
	m_clMiner.Kern.SetArg(14, m_bIsVector2 ? (UInt32(WorkData.FirstNonce) >> 1) : UInt32(WorkData.FirstNonce));
	m_clMiner.Kern.SetArg(15, m_w[2]);	
	m_clMiner.Kern.SetArg(16, m_w[16]);
	m_clMiner.Kern.SetArg(17, m_w[17]);
	m_clMiner.Kern.SetArg(18, m_preVal4);
	m_clMiner.Kern.SetArg(19, m_t1);

	m_clMiner.Kern.SetArg(20, Buffer);

	Cl::NDRange range(m_bIsVector2 ? m_nparOrig/2 : m_nparOrig);

#ifdef X_DEBUG//!!!D
	static int s_i;
	if (++s_i > 1) {
		m_clMiner.Queue.Enqueue(m_clMiner.Kern, range);
		m_ev = m_clMiner.Queue.EnqueueMarker();
		m_ev.SetCallback(&EventNotify, 0);
		m_clMiner.Queue.Flush();
		return;
	}
#endif

	m_clMiner.Queue.Enqueue(m_clMiner.Kern, range);
	m_clMiner.Queue.EnqueueRead(Buffer, m_bufTest, sizeof(m_bufTest));
	m_ev = m_clMiner.Queue.EnqueueMarker();

	//!!!ev.SetCallback(&EventNotify, 0);
	m_clMiner.Queue.Flush();
}

bool OpenclTask::IsComplete() {
#ifdef X_DEBUG//!!!D

	//Cl::Wait(m_ev);
	//m_ev.Close();
	Sleep(200);
	return true;
#endif
	Cl::ExecutionStatus st = m_ev.Status;
	if (st == Cl::ExecutionStatus::Complete) {
		return true;
	}
	int rc = (int)st;
	if (rc < 0)
		Cl::ClCheck(rc);
	return false;
}

void OpenclTask::Wait() {
	Cl::Wait(m_ev);
}

int OpenclTask::ParseResults() {
	int r = 0;
	if (m_bufTest[256]) {
		for (int i=0; i<256; ++i) {
			if (UInt32 nonce = m_bufTest[i]) {
				++r;
				if (m_pwt)
					m_pwt->Miner.TestAndSubmit(*m_pwt, WorkData, nonce);
			}
		}
		ZeroStruct(m_bufTest);
		m_clMiner.Queue.EnqueueWrite(Buffer, m_bufTest, sizeof(m_bufTest));
	}

	if (m_pwt)
		Interlocked::Add(m_pwt->Miner.HashCount, m_nparOrig/UCFG_BITCOIN_NPAR);
	return r;
}

ptr<GpuMiner> CreateOpenclMiner(BitcoinMiner& miner, Cl::Device& dev, const ConstBuf& binary) {
	return new OpenclMiner(miner, dev, binary);
}

OpenclMiner::OpenclMiner(BitcoinMiner& miner, const Cl::Device& dev, const ConstBuf& binary)
	:	base(miner)
	,	Device(dev)
{
	Cl::CContextProps props;
	props[CL_CONTEXT_PLATFORM] = (cl_context_properties)dev.get_Platform().Id;
	props[CL_CONTEXT_OFFLINE_DEVICES_AMD] = 1;
	Cl::Context ctx;
	Ctx.Create(props, dev);
	Prog = Cl::Program::FromBinary(Ctx, Device, binary);
	Prog.Build(Device);
	Kern = Prog.CreateKernel("search");		
}

/*!!!R void OpenclMiner::InitBeforeRun() {
	DevicePortion *= 16;
}*/

int OpenclMiner::Run(const BitcoinWorkData& wd, WorkerThreadBase *pwt) {

	/*!!!R
	while (true) {
		Cl::ExecutionStatus st = ev.Status;
		if (st == Cl::ExecutionStatus::Complete)
			break;
		int rc = st;
		if (rc < 0)
			Cl::ClCheck(rc);
		Sleep(1);
	}
	*/
	
	
	//Sleep(10000);
	//Wait(ev);


	//Wait(ev);

#ifdef X_DEBUG//!!!D
	memcpy(sha.m_w1, sha.m_midstate_after_3, 8*sizeof(UInt32));		
	sha.CalcRoundsTest(sha.m_w, sha.m_midstate, sha.m_w1, 16, 3, 4);
#endif

	return 0;
}

} // Coin::
