/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#pragma once

#include <el/libext/ext-http.h>
#include <el/inet/async-text-client.h>
#include <el/inet/json-rpc.h>
using namespace Ext::Inet;

#include "bitcoin-common.h"

#if UCFG_BITCOIN_USE_CAL
#	include <el/comp/gpu-cal.h>
using namespace Ext::Gpu;
#endif


#include "bitcoin-sha256.h"

#include "util.h"
#include "miner-interface.h"

namespace Ext {
	namespace Cl {
	class Device;
}}

namespace Coin {

class BitcoinMiner;

const float MAX_GPU_TEMPERATURE = 83; // Celsius
const float MAX_FPGA_TEMPERATURE = 60; // Celsius

const int BITCOIN_DEFAULT_GPU_IDLE_MILLISECONDS = 100;

const UInt32 FULL_TASK = UInt32(0x100000000ULL / UCFG_BITCOIN_NPAR);

const int MIN_GETWORK_QUEUE = 3,
			MAX_GETWORK_QUEUE = 5;

const int NORMAL_WAIT = 2000;

struct MinerTx {
	Blob Data;
	HashValue Hash;
};

class MinerBlock : public BlockBase {
	typedef BlockBase base;
public:
	VarValue JobId;
	mutable vector<MinerTx> Txes;
	String WorkId;
	Blob ConBaseAux;
	DateTime CurTime;
	TimeSpan ServerTimeOffset;
	DateTime MyExpireTime;
	Int64 CoinbaseValue;
	Blob Coinb1, Extranonce1, Extranonce2, CoinbaseAux, Coinb2;
	CCoinMerkleBranch MerkleBranch;

	byte TargetBE[32];

	MinerBlock()
		:	WorkId(nullptr)
		,	CoinbaseValue(-1)		
	{}

	static MinerBlock FromJson(const VarValue& json);
	static ptr<MinerBlock> FromStratumJson(const VarValue& json);
	Coin::HashValue MerkleRoot(bool bSave=true) const override;
protected:
	void SetTimestamps(const DateTime& dt) {
		ServerTimeOffset = (CurTime = dt) - DateTime::UtcNow();
	}
};

BinaryWriter& operator<<(BinaryWriter& wr, const MinerBlock& minerBlock);

class BitcoinWorkData {
	typedef BitcoinWorkData class_type;
public:
	DateTime Timestamp;
	Blob Midstate, Data, Hash1;
	byte TargetBE[32];
	UInt32 FirstNonce, LastNonce;
	UInt16 RollNTime;
	Coin::HashAlgo HashAlgo;

	ptr<Coin::MinerBlock> MinerBlock;

	BitcoinWorkData()
		:	FirstNonce(0)
		,	LastNonce(UInt32(-1))
		,	RollNTime(0)
	{}

	EXPLICIT_OPERATOR_BOOL() const {
		return Data.Size != 0 ? EXT_CONVERTIBLE_TO_TRUE : 0;
	}

	bool IsFull() const {
		return FirstNonce==0 && LastNonce == UInt32(-1);
	}

	static BitcoinWorkData FromJson(RCString sjson, Coin::HashAlgo algo);
	void SetRollNTime(WebClient& wc);

	UInt32 get_Nonce() const;
	void put_Nonce(UInt32 v);
	DEFPROP(UInt32, Nonce);

	bool TestNonceGivesZeroH(UInt32 nonce);
};

class ComputationDevice : public Object {
public:
	typedef Interlocked interlocked_policy;

	String Name, Description;
	Ext::Temperature Temperature;
	volatile Int32 HwErrors;
	CBool IsAmdGpu;
//!!!	int Index;
	ptr<Object> Miner;

	ComputationDevice()
		:	HwErrors(0)
	{}

	virtual void Start(BitcoinMiner& miner, CThreadRef *tr) =0;

	virtual Ext::Temperature GetTemperature() { return Temperature; }
};

class ComputationArchitecture {
public:
	static ComputationArchitecture *s_pChain;
	ComputationArchitecture *m_pNext;

	ComputationArchitecture()
		:	m_pNext(exchange(s_pChain, this))
	{
	}

	virtual void FillDevices(BitcoinMiner& miner) =0;
	virtual void DetectNewDevices(BitcoinMiner& miner) {}
};

class GpuArchitecture : public ComputationArchitecture {
};

class SerialDeviceArchitecture : public ComputationArchitecture {
public:
	void DetectNewDevices(BitcoinMiner& miner) override;
protected:
	DateTime m_dtNextDetect;
	String SysName;

	bool TimeToDetect();

};

class FpgaArchitecture : public SerialDeviceArchitecture {
protected:
};


class WorkerThreadBase : public Thread {
	typedef Thread base;
public:
	BitcoinMiner& Miner;
	Int64 m_prevTsc;
	UInt32 DevicePortion;
	ptr<ComputationDevice> Device;

	WorkerThreadBase(BitcoinMiner& miner, CThreadRef *tr)
		:	base(tr)
		,	Miner(miner)
		,	CurrentWebClient(0)
		,	m_nHashes(0)
		,	m_prevTsc(0)
		,	DevicePortion(128)
	{
		m_bSleeping = true;
	}

	virtual String GetDeviceName() { return nullptr; };
	void Stop() override;
protected:
	WebClient *CurrentWebClient;
	TimeSpan m_span;
	UInt64 m_nHashes;

#if UCFG_WIN32
	void OnAPC() override {
		base::OnAPC();
		if (CurrentWebClient)
			CurrentWebClient->CurrentRequest->ReleaseFromAPC();
	}
#endif

	void CompleteRound(UInt64 nHashes);

	void StartRound() {
		m_dtStartRound = DateTime::UtcNow();
	}
private:
	DateTime m_dtStartRound;

	friend class BitcoinMiner;
};

class CpuDevice : public ComputationDevice {
public:
	void Start(BitcoinMiner& miner, CThreadRef *tr) override;
};

class GpuDevice : public ComputationDevice {
};

class AmdGpuDevice : public GpuDevice {
public:
	AmdGpuDevice() {
		IsAmdGpu = true;
	}

	void Start(BitcoinMiner& miner, CThreadRef *tr) override;
};

class GpuThreadBase : public WorkerThreadBase {
	typedef WorkerThreadBase base;
public:
	GpuThreadBase(BitcoinMiner& miner, CThreadRef *tr)
		:	base(miner, tr)
	{}

};

struct SUrlTtr {
	String Url;
	int TtrMinutes;

	bool operator==(const SUrlTtr& x) const {
		return Url==x.Url && TtrMinutes==x.TtrMinutes;
	}

	bool operator!=(const SUrlTtr& x) const {
		return !operator==(x);
	}
};

class BitcoinWebClient : public WebClient {
	typedef WebClient base;

public:
	int Timeout;

	BitcoinWebClient()
		:	Timeout(-1)
	{}
protected:
	void OnHttpWebRequest(HttpWebRequest& req) {
		base::OnHttpWebRequest(req);
		if (Timeout >= 0)
			req.Timeout = Timeout;
	}
};


void ThrowRejectionError(RCString reason);

class StratumTask : public JsonRpcRequest {
public:
	int State;
	BitcoinWorkData m_wd;

	StratumTask(int state)
		:	State(0)
	{}
};

class StratumClient : public AsyncTextClient {
	typedef AsyncTextClient base;
public:
	BitcoinMiner& Miner;
	int State;
	Ext::Inet::JsonRpc JsonRpc;
	byte TargetBE[32];
	Blob Extranonce1, Extranonce2;

	StratumClient(Coin::BitcoinMiner& miner);
	void Call(RCString method, const vector<VarValue>& params, ptr<StratumTask> task = nullptr);
	void Submit(const BitcoinWorkData& wd);
protected:
	void SetDifficulty(double difficulty);
	void OnLine(RCString line) override;
	void Execute() override;
};

class MINER_CLASS BitcoinMiner : public IMiner {
public:
//	static const Int32 NONCE_STEP = 128*1024*1024; //!!! was 8
	
	int GetworkPeriod;
	float Speed;		// float to be atomary
	Int64 EntireHashCount;
	DateTime DtStart;

	mutex m_csGetWork;
	ManualResetEvent m_evDataReady;
	AutoResetEvent m_evGetWork;

	int GpuIdleMilliseconds;
	CBool m_bTryGpu, m_bTryFpga, m_bLongPolling;
	Coin::HashAlgo HashAlgo;

	byte BestHashBE[32];

	mutex MtxDevices;
	vector<ptr<ComputationDevice>> Devices;
	unordered_set<String> WorkingDevices, TestedDevices;

	ostream *m_pTraceStream;
	wostream *m_pWTraceStream;

	int m_msWait;
	int NPAR;
	volatile Int32 HashCount;
	volatile Int32 AcceptedCount;
	CInt<int> Intensity;
	int Verbosity;
	String UserAgentString;
	ptr<Coin::StratumClient> StratumClient;
	
	mutex m_csCurrentData;
	ptr<Coin::MinerBlock> MinerBlock;

	typedef list<BitcoinWorkData> CTaskQueue;
	CTaskQueue TaskQueue;

	list<BitcoinWorkData> WorksToSubmit;

	DateTime m_dtNextGetWork;

	String MainBitcoinUrl, Login, Password;
	DateTime m_dtSwichToMainUrl;
	String ProxyString;

	int ThreadCount;
	CPointer<CThreadRef> m_tr;
	vector<ptr<Thread> > Threads;
	ptr<Thread> GetWorkThread;
	ptr<Thread> GpuThread;
	int m_minGetworkQueue; 
	Temperature MaxGpuTemperature, MaxFpgaTemperature;

//!!!R	typedef LruMap<HashValue, Coin::MinerBlock> CLastBlockCache;
//!!!R	CLastBlockCache m_lastBlockCache;


	CBool m_bStop;
	CBool m_bInited;

	BitcoinMiner();
	virtual ~BitcoinMiner() {}

	void Release() override { delete this; }

	void InitDevices(vector<String>& selectedDevs);
	BitcoinWorkData GetWorkForThread(WorkerThreadBase& wt, UInt32 portion, bool bAllHashAlgoAllowed);
	virtual BitcoinWebClient GetWebClient(WorkerThreadBase *wt);
//!!!	void SetNewData(const BitcoinWorkData& wd);
	void SetNewData(const BitcoinWorkData& wd, bool bClearOld = false);
	void SetWebInfo(WebClient& wc);
	void CheckLongPolling(WebClient& wc, RCString longPollUri = nullptr, RCString longPollId = nullptr);
	void CallNotifyRequest(WorkerThreadBase& wt, RCString url, RCString longPollId);	
	
	String GetMethodName(bool bSubmit);
	BitcoinWorkData GetWorkFromMinerBlock(const DateTime& now);
	virtual BitcoinWorkData GetWork(WebClient*& curWebClient);
	virtual bool SubmitResult(WebClient*& curWebClient, const BitcoinWorkData& wd);
	
	void SetIntensity(int v) override {
		Intensity = v;
	}

	Blob CalcHash(const BitcoinWorkData& wd);
	bool TestAndSubmit(WorkerThreadBase& wt, BitcoinWorkData& wd, UInt32 nonce);
	virtual String GetCalIlCode(bool bEvergreen);
	virtual String GetOpenclCode();
	virtual String GetCudaCode();

	void Print(const BitcoinWorkData& wd, bool bSuccess, RCString message = "");

	bool UseSse2() {
#if UCFG_CPU_X86_X64
		return CpuInfo().HasSse2;
#endif
		return false;
	}

	void CreateStratumClient(RCString s);

	void Start(CThreadRef *tr);
	void Start() override { Start(nullptr); }

	virtual void OnRoundComplete() {}

	void Stop() override;

	void SetMainUrl(const char *url) override { MainBitcoinUrl = url; }
	void SetLogin(const char *login) override { Login = login; }
	void SetPassword(const char *password) override { Password = password; }
	void SetThreadCount(int n) override { ThreadCount = n; }
	void SetHashAlgo(int a) override { HashAlgo = (Coin::HashAlgo)a; }
	double GetSpeed() override { return Speed; }

	BitcoinWorkData GetTestData();
private:
	CInt<int> m_getWorkMethodLevel;

	String BitcoinUrl;
	vector<SUrlTtr> HostList;
	CInt<int> HostIndex;

	String GetCurrentUrl();

	mutex m_csLongPollingThread;
	ptr<Thread> m_LongPollingThread;

	friend class WorkerThread;
	friend class LongPollingThread;
	friend class GetWorkThread;
};

class GpuMiner;

class GpuTask : public BitcoinSha256 {
public:
	BitcoinWorkData WorkData;
	WorkerThreadBase *m_pwt;
	int m_npar, m_nparOrig;
	GpuMiner& m_gpuMiner;
	CBool m_bIsVector2;

	GpuTask(GpuMiner& gpuMiner)
		:	m_gpuMiner(gpuMiner)
	{}

	virtual void Prepare(const BitcoinWorkData& wd, WorkerThreadBase *pwt) =0;
	virtual void Run() =0;
	virtual bool IsComplete() =0;
	virtual void Wait() =0;
	virtual int ParseResults() =0;
};

class GpuMiner : public Object {
public:
	BitcoinMiner& Miner;
	UInt32 DevicePortion;
	CBool m_bIsVector2;

	GpuMiner(BitcoinMiner& miner)
		:	Miner(miner)
		,	DevicePortion(1024*1024/UCFG_BITCOIN_NPAR)
	{}

//!!!	virtual void InitBeforeRun() {}
	virtual ptr<GpuTask> CreateTask() =0;
	virtual int Run(const BitcoinWorkData& wd, WorkerThreadBase *pwt) =0;
};

ptr<GpuMiner> CreateOpenclMiner(BitcoinMiner& miner, Cl::Device& dev, const ConstBuf& binary);

#if UCFG_BITCOIN_USE_CAL
bool StartCal();
ptr<GpuMiner> CreateCalMiner(BitcoinMiner& miner, CALtarget calTarget, int idx);
String TargetToString(CALtarget target);

struct CalEngineWrap {
	CalEngine Engine;

	static mutex s_cs;
	static volatile Int32 RefCount;
	static CalEngineWrap *I;
	
	static CalEngineWrap *Get() {
		EXT_LOCK (s_cs) {
			if (!I)
				I = new CalEngineWrap;
			return I;
		}
	}

	static void AddRef(CalEngineWrap *p) {
		Interlocked::Increment(RefCount);
	}

	static void Release(CalEngineWrap *p) {
		EXT_LOCK (s_cs) {
			if (!Interlocked::Decrement(RefCount)) {
				delete I;
				I = nullptr;
			}
		}
	}
};


#endif // UCFG_BITCOIN_USE_CAL


} // Coin::

