/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#pragma once

#ifdef __APPLE__
#	include <OpenCL/opencl.h>
#else
#	include <CL/cl.h>
#	include <CL/cl_ext.h>
#endif

#ifndef CL_VERSION_1_2
#	error OpenCL 1.2 or later version is required
#endif

namespace Ext { namespace Cl {

using namespace Ext;

class Device;
class Kernel;

void ClCheck(cl_int rc);


class Platform {
	typedef Platform class_type;
public:
	cl_platform_id Id;

	Platform(cl_platform_id id = 0)
		:	Id(id)
	{}

	static vector<Platform> GetAll();

	String get_Vendor() const;
	DEFPROP_GET(String, Vendor);

	vector<String> get_Extensions() const;
	DEFPROP_GET(vector<String>, Extensions);

	vector<Device> GetDevices(cl_device_type typ = CL_DEVICE_TYPE_ALL) const;
};

class Device {
	typedef Device class_type;
public:
	cl_device_id Id;

	Device(cl_device_id id = 0)
		:	Id(id)
	{}

	Blob GetInfo(cl_device_info pname) const;

	Cl::Platform get_Platform() const {
		Blob blob = GetInfo(CL_DEVICE_PLATFORM);
		return Cl::Platform(*(cl_platform_id*)blob.constData());
	}
	DEFPROP_GET(Cl::Platform, Platform);

	String get_Name() const {
		Blob blob = GetInfo(CL_DEVICE_NAME);
		return String((const char*)blob.constData());
	}
	DEFPROP_GET(String, Name);

	bool get_Available() const {
		Blob blob = GetInfo(CL_DEVICE_AVAILABLE);
		return *(cl_bool*)blob.constData();
	}
	DEFPROP_GET(bool, Available);

	UInt32 get_MaxComputeUnits() const {
		Blob blob = GetInfo(CL_DEVICE_MAX_COMPUTE_UNITS);
		return *(cl_uint*)blob.constData();
	}
	DEFPROP_GET(UInt32, MaxComputeUnits);

	size_t get_MaxWorkGroupSize() const {
		Blob blob = GetInfo(CL_DEVICE_MAX_WORK_GROUP_SIZE);
		return *(size_t*)blob.constData();
	}
	DEFPROP_GET(size_t, MaxWorkGroupSize);

	vector<String> get_Extensions() const {
		Blob blob = GetInfo(CL_DEVICE_EXTENSIONS);
		return String((const char*)blob.constData(), blob.Size).Split();
	}
	DEFPROP_GET(vector<String>, Extensions);
};


} // Cl::

#define DEF_CL_OBJECT(uname, lname) template<> struct HandleTraits<cl_##lname> : HandleTraitsBase<cl_##lname> {	\
	static handle_type Retain(const handle_type& h) { Cl::ClCheck(::clRetain##uname(h)); return h; }	\
	static void ResourceRelease(const handle_type& h) { Cl::ClCheck(::clRelease##uname(h)); }	\
};

DEF_CL_OBJECT(Context, context);
DEF_CL_OBJECT(Program, program);
DEF_CL_OBJECT(CommandQueue, command_queue);
DEF_CL_OBJECT(Kernel, kernel);
DEF_CL_OBJECT(Event, event);
DEF_CL_OBJECT(MemObject, mem);

/*!!!R

template<> struct HandleTraits<cl_context> : HandleTraitsBase<cl_context> {
	static cl_context Retain(const cl_context& h) { Cl::ClCheck(::clRetainContext(h)); return h; }
	static void ResourceRelease(const cl_context& h) { Cl::ClCheck(::clReleaseContext(h)); }
};

template<> struct HandleTraits<cl_program> : HandleTraitsBase<cl_program> {
	static cl_program Retain(const cl_program& h) { Cl::ClCheck(::clRetainProgram(h)); return h; }
	static void ResourceRelease(const cl_program& h) { Cl::ClCheck(::clReleaseProgram(h)); }
};

template<> struct HandleTraits<cl_command_queue> : HandleTraitsBase<cl_command_queue> {
	static cl_command_queue Retain(const cl_command_queue& h) { Cl::ClCheck(::clRetainCommandQueue(h)); return h; }
	static void ResourceRelease(const cl_command_queue& h) { Cl::ClCheck(::clReleaseCommandQueue(h)); }
};

template<> struct HandleTraits<cl_kernel> : HandleTraitsBase<cl_kernel> {
	static cl_kernel Retain(const cl_kernel& h) { Cl::ClCheck(::clRetainKernel(h)); return h; }
	static void ResourceRelease(const cl_kernel& h) { Cl::ClCheck(::clReleaseKernel(h)); }
};*/

namespace Cl {

class NDRange {
public:
	static const int DEFAULT_WORK_GROUP_SIZE = 64;			//!!!?

	size_t m_global[3], m_local[3];
	size_t m_globalSize, m_localSize;

	NDRange(int n = 0);
};

ENUM_CLASS(ExecutionStatus) {
	Complete = CL_COMPLETE,
	Running = CL_RUNNING,
	Submitted = CL_SUBMITTED,
	Queued = CL_QUEUED
} END_ENUM_CLASS(ExecutionStatus);

class Event : public ResourceWrapper<cl_event> {
	typedef Event class_type;
public:
	void SetCallback(void (CL_CALLBACK *pfn)(cl_event, cl_int, void *), void *userData);

	ExecutionStatus get_Status();
	DEFPROP_GET(ExecutionStatus, Status);

	int get_RefCount();
	DEFPROP_GET(int, RefCount);

	friend void Wait(const Event& ev);
};

void Wait(const Event& ev);

class Buffer : public ResourceWrapper<cl_mem> {
};

class CommandQueue : public ResourceWrapper<cl_command_queue> {
public:
	Event Enqueue(const Kernel& kernel, const NDRange& range);
	Event Enqueue(const Kernel& kernel);
	Event EnqueueRead(const Buffer& buffer, void *p, size_t cb);
	Event EnqueueWrite(const Buffer& buffer, const void *p, size_t cb);
	Event EnqueueMarker();
	Event EnqueueBarrier();
	
	void Flush() {
		ClCheck(::clFlush(_self));
	}

	void Finish() {
		ClCheck(::clFinish(_self));
	}
};

typedef map<cl_context_properties, cl_context_properties> CContextProps;

class Context : public ResourceWrapper<cl_context> {
	typedef Context class_type;
public:
	~Context();
	void Create(const CContextProps& props, const Device& dev);
	void CreateFromType(const CContextProps& props, cl_device_type typ = CL_DEVICE_TYPE_ALL);
	CommandQueue CreateCommandQueue(const Device& dev, cl_command_queue_properties prop = 0);
	Buffer CreateBuffer(void *p, size_t size, cl_mem_flags flags = 0);

	vector<Device> get_Devices() const;
	DEFPROP_GET(vector<Device>, Devices);
};


struct ProgramBinary {
	cl_device_id DeviceId;
	Blob Binary;
};

class Kernel : public ResourceWrapper<cl_kernel> {
public:
	void SetArg(cl_uint idx, const Buffer& buf) {
		ClCheck(::clSetKernelArg(_self, idx, sizeof(buf.m_h), &buf.m_h));
	}

	template <typename T>
	void SetArg(cl_uint idx, const T& v) {
		ClCheck(::clSetKernelArg(_self, idx, sizeof(v), &v));
	}
};

class Program : public ResourceWrapper<cl_program> {
	typedef Program class_type;
public:
	CPointer<Context> Ctx;

	typedef unordered_map<String, String> COptions;
	COptions Options;

	void CreateFromSources(Context& ctx, const vector<String>& ss);
	static Program FromBinary(Context& ctx, const Device& dev, const ConstBuf& mb);
	void Build(Device& dev);
	Kernel CreateKernel(RCString name);	

	vector<ProgramBinary> get_Binaries();
	DEFPROP_GET(vector<ProgramBinary>, Binaries);
};


class OpenclException : public Exc {
	typedef Exc base;
public:
	OpenclException(HRESULT hr)
		:	base(hr)
	{}
};

class BuildException : public OpenclException {
	typedef OpenclException base;
public:
	String Log;

	BuildException(HRESULT hr, RCString log)
		:	base(hr)
		,	Log(log)
	{}
};

}} // Ext::Cl::

